/* 
 * Graph Bisect Library
 *
 * Grid Challenge 2006 Final
 * Team Leo - Nan DUN
 * Department of Computer Science, the University of Tokyo
 */

#ifndef GB_H
#define GB_H

/*******************************************
 * Constants, types
 *******************************************/
#define GB_MAX_FILENAME	128

#define GB_PAR_N	0
#define GB_PAR_L	1
#define GB_PAR_R	2

#define GB_NUM_EVEN	0
#define GB_NUM_ODD	1

#define GB_FALSE	0
#define GB_TRUE		1

typedef	int integer;

typedef struct GB_GRAPH {
	char filename[GB_MAX_FILENAME];
	integer nv;			/* Number of vertices */
	integer ne;			/* Number of edges */
	integer **adjlist;	/* Vertices adjacent list, adjlist[vertex][0] 
						   is the number of adjacent vertices to vertex. */
	int vparity;		/* Parity of vertices number */
	integer maxvdeg;	/* Maximum vertex degree */
	integer minvdeg;	/* Minimum vertex degree */
	
	integer *belonglist;/* Vertices belonging list: GB_PAR_L or GB_PAR_R */
	integer	edgecut;	/* Edge cut number of bisection */
} GB_GRAPH;

typedef enum GB_CONV {
	GB_M2J, 	/* Metis to Johnson */
	GB_DB2J		/* DIMACS binary to Johnson */
} GB_CONV;

typedef enum GB_ALOG {
	GB_KL,		/* Kernighan-Lin */
	GB_LKL,
	GB_GG,		/* Greedy growing */
	GB_MMG,		/* Min-Max greedy */
	GB_RTS,		/* Randomized tabu search */
	GB_RRTS		/* Reactive randomized tabu search */
} GB_ALGO;

/*******************************************
 * APIs
 *******************************************/
void GB_Convert(const char *filefrom, const char *fileto, GB_CONV options);
void GB_Initial(const char *filename, GB_GRAPH *graph);
void GB_Finalize(GB_GRAPH *graph);
void GB_SeedDistribute(int hostnum, int hostid, integer nv, int *seed);
void GB_Bisect(GB_GRAPH *graph, GB_ALGO options, integer *seed, int iter, double timeout, integer threshold);
void GB_Check(GB_GRAPH *graph);

#endif	/* GB_H */

