/* 
 * Graph Bisect Library
 * Implementation of APIs
 *
 * Grid Challenge 2006 Final
 * Team Leo - Nan DUN
 * Department of Computer Science, the University of Tokyo
 */

#include "gb.h"
#include "gbimpl.h"

void GB_Convert(const char *filefrom, const char *fileto, GB_CONV options) {
	const char FUNCNAME[] = "GB_Convert(3)";

	if (filefrom == NULL || fileto == NULL) {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}

	switch (options) {
		case GB_M2J:
			GB_Metis2Johnson(filefrom, fileto);
			break;

		case GB_DB2J:
			GB_DimacsB2Johnson(filefrom, fileto);
			break;

		default:
			GB_Error(GB_ERR_ARG, FUNCNAME);
			break;
	}
}

void GB_Initial(const char *filename, GB_GRAPH *graph) {
	const char FUNCNAME[] = "GB_Initial(2)";

	if (graph == NULL) {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}
	GB_ReadGraph(filename, graph);
}

void GB_Finalize(GB_GRAPH *graph) {
	const char FUNCNAME[] = "GB_Finalized(1)";

	if (graph == NULL) {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}
	g_io_start = clock();
	GB_FreeGraph(graph);
	g_io_end = clock();
}

void GB_Check(GB_GRAPH *graph) {
	const char FUNCNAME[] = "GB_Finalized(1)";

	if (graph == NULL) {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}
	GB_PrintGraph(graph);
}

void GB_Bisect(GB_GRAPH *graph, GB_ALGO options, integer *seed, int iter, double timeout, integer threshold) {
	GB_GlobalInitial(graph, options, seed, iter, timeout, threshold);

	g_bi_start = clock();

	switch (options) {
		case GB_MMG:
			GB_MinMaxGreedyBisect();
			break;
		case GB_RTS:
			GB_RandomizedTabuSearchBisect();
			break;
		case GB_RRTS:
			GB_ReactiveRandomizedTSBisect();
			break;
		case GB_GG:
			GB_GreedyGrowingRegionBisect();
			break;
		case GB_KL:
			GB_KernighanLinBisect();
			break;
		case GB_LKL:
			GB_LinearKernighanLinBisect();
			break;
		default:
			break;
	}

	g_bi_end = clock();

	GB_GlobalFinalized();
}

