/* 
 * Graph Bisect Library
 * Greedy Growing 
 *
 * Grid Challenge 2006 Final
 * Team Leo - Nan DUN
 * Department of Computer Science, the University of Tokyo
 */

#include "gbimpl.h"

integer *g_vcutlist;
extern bucket_t *g_gainBuckets;
extern integer g_gainMin[3], g_gainMax[3], g_gainBktsOffset;
extern itemp_t *g_vgainlist;

extern void GainBucketReset(bucket_t *buckets, integer ng, itemp_t *vlist, integer ni);
integer GainBucketDetachMin(int set, bucket_t *buckets, integer *gainMin, integer *gainMax, integer offset, itemp_t *vlist, integer *gain);
void UpdateBoundary(integer vertex, integer *boundary);

extern void GainBucketInsert(int set, integer gain, integer vertex, bucket_t *buckets, integer *gainMin, integer *gainMax, integer offset, itemp_t *vlist);
extern void GainBucketUpdate(int oldset, int newset, integer oldgain, integer newgain, integer vertex, bucket_t *buckets, integer *gainMin, integer *gainMax, integer offset, itemp_t *vlist);

void GB_GreedyGrowingRegionBisect() {
	integer i, j;

	g_timelimit = g_timeout + g_bi_start;
	for (i = 0; i < g_iter; i++) {
		for (j = 1; j <= *g_nv; j++) {
			(*g_belonglist)[j] = GB_PAR_L;
		}
		GB_GreedyGrowingRegionBisectSub();
		if (g_exitflag == GB_TRUE) {
			goto FOUND_EXIT;
		}
	}
FOUND_EXIT:
	*g_edgecut = g_bestedgecut;
	for (i = 1; i <= *g_nv; i++) {
		(*g_belonglist)[i] = g_bestbelonglist[i]; 
	}
}

void GB_GreedyGrowingRegionBisectSub() {
	integer tobeadded, minvertex, gain, i, j, temp;

	*g_edgecut = 0;
	tobeadded = *g_nv / 2;
	GB_RandomPermute(g_randomvlist, *g_nv);
	for (i = 1; i <= *g_nv; i++) {
		g_vcutlist[i] = 0;
	}
	GainBucketReset(g_gainBuckets, 2 * (*g_maxvdeg) + 1, g_vgainlist, *g_nv + 1);
	g_gainMin[0] = *g_maxvdeg;
	g_gainMax[0] = 0 - *g_maxvdeg;
	i = 1;
	while (tobeadded > 0) {
		if (g_gainMin[0] > g_gainMax[0]) {	/* Empty list */
			minvertex = g_randomvlist[i];
			gain = (*g_adjlist)[minvertex][0];
			GainBucketInsert(0, gain, minvertex, g_gainBuckets, g_gainMin, g_gainMax, g_gainBktsOffset, g_vgainlist);
			i++;
		}
		minvertex = GainBucketDetachMin(0, g_gainBuckets, g_gainMin, g_gainMax, g_gainBktsOffset, g_vgainlist, &gain);
		(*g_belonglist)[minvertex] = GB_PAR_R;
		tobeadded--;
		*g_edgecut += gain;
		UpdateBoundary(minvertex, &(g_vcutlist[minvertex]));

		if (g_vcutlist[minvertex] > 0) {
			for (i = 1; i <= (*g_adjlist)[minvertex][0]; i++) {
				temp = (*g_adjlist)[minvertex][i];
				if ((*g_belonglist)[temp] == GB_PAR_L) {
					if (g_vcutlist[temp] == 0) {
						g_vcutlist[temp] = -1;
						gain = g_vgainlist[temp].value = ((*g_adjlist)[temp][0]) - 2;
						GainBucketInsert(0, gain, temp, g_gainBuckets, g_gainMin, g_gainMax, g_gainBktsOffset, g_vgainlist);
					} else {
						gain = g_vgainlist[temp].value - g_gainBktsOffset;
						GainBucketUpdate(0, 0, gain, gain - 2, temp, g_gainBuckets, g_gainMin, g_gainMax, g_gainBktsOffset, g_vgainlist);
					}
				}
			}
		}

	}
	if (*g_edgecut < g_bestedgecut) {
		g_bestedgecut = *g_edgecut;
		for (j = 1; j <= *g_nv; j++) {
			g_bestbelonglist[j] = (*g_belonglist)[j];
		}
	}
	if (g_bestedgecut <= g_threshold || (g_timeout > 0 && clock() >= g_timelimit)) {
		g_exitflag = GB_TRUE;
	}
}

void UpdateBoundary(integer vertex, integer *boundary) {
	integer i;
	integer *adjlist, *belonglist;

	adjlist = (*g_adjlist)[vertex];
	belonglist = *g_belonglist;

	*boundary = 0;
	for (i = 1; i <= adjlist[0]; i++) {
		if (belonglist[adjlist[i]] == GB_PAR_L) {
			(*boundary)++;
		}
	}
}

integer GainBucketDetachMin(int set, bucket_t *buckets, integer *gainMin, integer *gainMax, integer offset, itemp_t *vlist, integer *gain) {
	integer vertex, mymin, mymax;

	mymin = gainMin[set] + offset;
	mymax = gainMax[set] + offset;

	vertex = GB_BucketGetFirstMin(g_gainBuckets, mymin);
	*gain = vlist[vertex].value - offset;
	GB_BucketDelete(mymin, vertex, buckets, &mymin, &mymax, vlist);
	if (mymin > mymax) {
		gainMin[set] = *g_maxvdeg;
		gainMax[set] = 0 - *g_maxvdeg;
	} else {
		gainMin[set] = mymin - offset;
		gainMax[set] = mymax - offset;
	}
	return vertex;
}

