/* 
 * Graph Bisect Library
 * Kernighan-Lin
 *
 * Grid Challenge 2006 Final
 * Team Leo - Nan DUN
 * Department of Computer Science, the University of Tokyo
 */

#include "gbimpl.h"

#define UNLOCK	0
#define LOCK	1

integer *g_dvaluelist, *g_edgecutlist; /* d-value is actually gain */
int *g_vlocklist;
pair_t *g_vpairlist;

void RandomBisection();
integer Omega(integer a, integer b);

void GB_KernighanLinBisect() {
	int i;
	clock_t timeout;

	timeout = g_timeout + g_bi_start;
	RandomBisection();
	for (i = 0; i < g_iter; i++) {
		GB_KernighanLinBisectSub();		
		if (g_exitflag == GB_TRUE) {
			goto FOUND_EXIT;
		}
	}
FOUND_EXIT:
	*g_edgecut = g_bestedgecut;
	for (i = 1; i <= *g_nv; i++) {
		(*g_belonglist)[i] = g_bestbelonglist[i]; 
	}
}

void GB_KernighanLinBisectSub() {
	integer myset, internal, external, bestedgecut, bestchange;
	pair_t minpair, temppair;
	integer i, j, k, n_2, tempvertex, tempvalue;

	/* compute gain values */
	for (i = 1; i <= *g_nv; i++) {
		myset = (*g_belonglist)[i];
		internal = external = 0;
		for (j = 1; j <= (*g_adjlist)[i][0]; j++) {
			tempvertex = (*g_adjlist)[i][j];
			if (myset == (*g_belonglist)[tempvertex]) {
				internal++;
			} else {
				external++;
			}
		}
		g_dvaluelist[i] = external - internal;
		g_vlocklist[i] = UNLOCK;
	}
	
	bestedgecut = g_edgecutlist[0] = *g_edgecut;
	if (bestedgecut == 0) {
		return;
	}
	bestchange = 0;
	n_2 = *g_nv / 2;

	for (i = 1; i <= n_2; i++) {
		g_edgecutlist[i] = *g_ne;
		for (j = 1; j <= *g_nv; j++) {
			if ((*g_belonglist)[j] == GB_PAR_L && g_vlocklist[j] == UNLOCK) {
				for (k = 1; k <= *g_nv; k++) {
					if ((*g_belonglist)[k] == GB_PAR_R && g_vlocklist[k] == UNLOCK) {
						tempvalue = 2 * Omega(j, k) - g_dvaluelist[j] - g_dvaluelist[k];
						if (tempvalue < g_edgecutlist[i]) {
							g_vpairlist[i].a = j;
							g_vpairlist[i].b = k;
							g_edgecutlist[i] = tempvalue;
						}
					}
				}
			}
		}

		minpair = g_vpairlist[i];
		g_vlocklist[minpair.a] = g_vlocklist[minpair.b] = LOCK;
		for (j = 1; j <= *g_nv; j++) {
			if (g_vlocklist[j] == UNLOCK) {
				if ((*g_belonglist)[j] == GB_PAR_L) {
					g_dvaluelist[j] = g_dvaluelist[j] - 2 * (Omega(j, minpair.b) - Omega(j, minpair.a));
				}else {
					g_dvaluelist[j] = g_dvaluelist[j] - 2 * (Omega(j, minpair.a) - Omega(j, minpair.b));
				}
			}
		}
		g_edgecutlist[i] += g_edgecutlist[i - 1];
		if (g_edgecutlist[i] < bestedgecut) {
			bestchange = i;
			bestedgecut = g_edgecutlist[i];
		}
	}
	for(i = 1; i <= bestchange; i++) {
		temppair = g_vpairlist[i];
		if ((*g_belonglist)[temppair.a] == GB_PAR_L) {
			(*g_belonglist)[temppair.a] = GB_PAR_R;
			(*g_belonglist)[temppair.b] = GB_PAR_L;
		} else {
			(*g_belonglist)[temppair.a] = GB_PAR_L;
			(*g_belonglist)[temppair.b] = GB_PAR_R;
		}
	}
	*g_edgecut = bestedgecut;
	if (g_bestedgecut > *g_edgecut) {
		g_bestedgecut = *g_edgecut;
		for (i = 1; i <= *g_nv; i++) {
			g_bestbelonglist[i] = (*g_belonglist)[i];
		}
	}
	if (g_bestedgecut <= g_threshold || (g_timeout > 0 && clock() >= g_timelimit)) {
		g_exitflag = GB_TRUE;
	}
}

void RandomBisection() {
	integer i, j, n_2;
	n_2 = *g_nv / 2;
	GB_RandomPermute(g_randomvlist, *g_nv);
	for (i = 1; i <= *g_nv; i++) {
		if (i <= n_2) {
			(*g_belonglist)[g_randomvlist[i]] = GB_PAR_L;
		} else {
			(*g_belonglist)[g_randomvlist[i]] = GB_PAR_R;
		}
	}
	*g_edgecut = 0;
	for (i = 1; i <= *g_nv; i++) {
		if ((*g_belonglist)[i] == GB_PAR_L) {
			for (j = 1; j <= (*g_adjlist)[i][0]; j++) {
				if ((*g_belonglist)[((*g_adjlist)[i][j])] == GB_PAR_R)
					(*g_edgecut)++;
			}
		}
	}
}
integer Omega(integer a, integer b) {
	integer i;

	for (i = 1; i <= (*g_adjlist)[a][0]; i++) {
		if (b == (*g_adjlist)[a][i]) {
			return 1;
		}
	}
	return 0;
}

