/* 
 * Graph Bisect Library
 * Kernighan-Lin
 *
 * Grid Challenge 2006 Final
 * Team Leo - Nan DUN
 * Department of Computer Science, the University of Tokyo
 */

#include "gbimpl.h"

#define UNLOCK	0
#define LOCK	1

extern integer *g_edgecutlist;
extern pair_t *g_vpairlist;
extern int *g_vlocklist;

extern void RandomBisection();
extern integer Omega(integer a, integer b);

bucket_t *g_gainBkt[3];
integer g_gainBktMin[3], g_gainBktMax[3], g_gainBktOffset;
itemp_t *g_gainlst[3];

extern void GainBucketReset(bucket_t *buckets, integer ng, itemp_t *vlist, integer ni);
extern void GainBucketInsert(int set, integer gain, integer vertex, bucket_t *buckets, integer *gainMin, integer *gainMax, integer offset, itemp_t *vlist);
extern void GainBucketDelete(int set, integer gain, integer vertex, bucket_t *buckets, integer *gainMin, integer *gainMax, integer offset, itemp_t *vlist);
extern void GainBucketUpdate(int oldset, int newset, integer oldgain, integer newgain, integer vertex, bucket_t *buckets, integer *gainMin, integer *gainMax, integer offset, itemp_t *vlist);


void GB_LinearKernighanLinBisect() {
	int i;

	g_timelimit = g_timeout + g_bi_start;
	RandomBisection();
	for (i = 0; i < g_iter; i++) {
		GB_LinearKernighanLinBisectSub();		
		if (g_exitflag == GB_TRUE) {
			goto FOUND_EXIT;
		}
	}
FOUND_EXIT:
	*g_edgecut = g_bestedgecut;
	for (i = 1; i <= *g_nv; i++) {
		(*g_belonglist)[i] = g_bestbelonglist[i]; 
	}
}

void GB_LinearKernighanLinBisectSub() {
	integer myset, internal, external, gain, bestedgecut, bestchange;
	pair_t minpair, temppair;
	integer i, j, n_2, vL, vR, gainL, gainR, oldgain, newgain, tempvertex, tempvalue;

	GainBucketReset(g_gainBkt[GB_PAR_L], 2 * (*g_maxvdeg) + 1, g_gainlst[GB_PAR_L], *g_nv + 1);
	GainBucketReset(g_gainBkt[GB_PAR_R], 2 * (*g_maxvdeg) + 1, g_gainlst[GB_PAR_R], *g_nv + 1);
	g_gainBktMin[GB_PAR_L] = g_gainBktMin[GB_PAR_R] = *g_maxvdeg;
	g_gainBktMax[GB_PAR_L] = g_gainBktMax[GB_PAR_R] = 0 - *g_maxvdeg;
	
	/* compute gain values */
	for (i = 1; i <= *g_nv; i++) {
		myset = (*g_belonglist)[i];
		internal = external = 0;
		for (j = 1; j <= (*g_adjlist)[i][0]; j++) {
			tempvertex = (*g_adjlist)[i][j];
			if (myset == (*g_belonglist)[tempvertex]) {
				internal++;
			} else {
				external++;
			}
		}
		gain = external - internal;
        GainBucketInsert(myset, gain, i, g_gainBkt[myset], g_gainBktMin, g_gainBktMax, g_gainBktOffset, g_gainlst[myset]);
		g_vlocklist[i] = UNLOCK;
	}

	bestedgecut = g_edgecutlist[0] = *g_edgecut;
	if (bestedgecut == 0) {
		return;
	}

	bestchange = 0;
	n_2 = *g_nv / 2;

	for (i = 1; i <= n_2; i++) {
		g_edgecutlist[i] = *g_ne;
		gainL = g_gainBktMax[GB_PAR_L];
		gainR = g_gainBktMax[GB_PAR_R];
		if (*g_vparity == GB_NUM_ODD && (gainL == -*g_maxvdeg || gainR == -*g_maxvdeg)) {
			return;
		}
		vL = (g_gainBkt[GB_PAR_L])[gainL+g_gainBktOffset].head->index;
		vR = (g_gainBkt[GB_PAR_R])[gainR+g_gainBktOffset].head->index;
        tempvalue = 2 * Omega(vL, vR) - gainL - gainR;
		if (tempvalue < g_edgecutlist[i]) {
			g_vpairlist[i].a = vL;
			g_vpairlist[i].b = vR;
			g_edgecutlist[i] = tempvalue;
		}
		minpair = g_vpairlist[i];
		GainBucketDelete(GB_PAR_L, gainL, vL, g_gainBkt[GB_PAR_L], g_gainBktMin, g_gainBktMax, g_gainBktOffset, g_gainlst[GB_PAR_L]);
		GainBucketDelete(GB_PAR_R, gainR, vR, g_gainBkt[GB_PAR_R], g_gainBktMin, g_gainBktMax, g_gainBktOffset, g_gainlst[GB_PAR_R]);
		g_vlocklist[vL] = 	g_vlocklist[vR] = LOCK;
		
		for (j = 1; j <= (*g_adjlist)[vL][0]; j++) {
			tempvertex = (*g_adjlist)[vL][j];
			if (g_vlocklist[tempvertex] == UNLOCK) {
				if ((*g_belonglist)[tempvertex] == GB_PAR_L) {
					oldgain = g_gainlst[GB_PAR_L][tempvertex].value - g_gainBktOffset;
					newgain = oldgain + 2;
					GainBucketUpdate(GB_PAR_L, GB_PAR_L, oldgain, newgain, tempvertex, g_gainBkt[GB_PAR_L], g_gainBktMin, g_gainBktMax, g_gainBktOffset, g_gainlst[GB_PAR_L]);
				}else {
					oldgain = g_gainlst[GB_PAR_R][tempvertex].value - g_gainBktOffset;
					newgain = oldgain - 2;
					GainBucketUpdate(GB_PAR_R, GB_PAR_R, oldgain, newgain, tempvertex, g_gainBkt[GB_PAR_R], g_gainBktMin, g_gainBktMax, g_gainBktOffset, g_gainlst[GB_PAR_R]);
				}
			}
		}
		for (j = 1; j <= (*g_adjlist)[vR][0]; j++) {
			tempvertex = (*g_adjlist)[vR][j];
			if (g_vlocklist[tempvertex] == UNLOCK) {
				if ((*g_belonglist)[tempvertex] == GB_PAR_R) {
					oldgain = g_gainlst[GB_PAR_R][tempvertex].value - g_gainBktOffset;
					newgain = oldgain + 2;
					GainBucketUpdate(GB_PAR_R, GB_PAR_R, oldgain, newgain, tempvertex, g_gainBkt[GB_PAR_R], g_gainBktMin, g_gainBktMax, g_gainBktOffset, g_gainlst[GB_PAR_R]);
				}else {
					oldgain = g_gainlst[GB_PAR_L][tempvertex].value - g_gainBktOffset;
					newgain = oldgain - 2;
					GainBucketUpdate(GB_PAR_L, GB_PAR_L, oldgain, newgain, tempvertex, g_gainBkt[GB_PAR_L], g_gainBktMin, g_gainBktMax, g_gainBktOffset, g_gainlst[GB_PAR_L]);
				}
			}
		}

		g_edgecutlist[i] += g_edgecutlist[i - 1];
		if (g_edgecutlist[i] < bestedgecut) {
			bestchange = i;
			bestedgecut = g_edgecutlist[i];
		}
	}

	for(i = 1; i <= bestchange; i++) {
		temppair = g_vpairlist[i];
		(*g_belonglist)[temppair.a] = GB_PAR_R;
		(*g_belonglist)[temppair.b] = GB_PAR_L;
	}
	*g_edgecut = bestedgecut;
	if (g_bestedgecut > *g_edgecut) {
		g_bestedgecut = *g_edgecut;
		for (i = 1; i <= *g_nv; i++) {
			g_bestbelonglist[i] = (*g_belonglist)[i];
		}
	}
	if (g_bestedgecut <= g_threshold || (g_timeout > 0 && clock() >= g_timelimit)) {
		g_exitflag = GB_TRUE;
	}
}

