/* 
 * Graph Bisect Library
 * Min-Max Greedy Bisect Routines
 *
 * Grid Challenge 2006 Final
 * Team Leo - Nan DUN
 * Department of Computer Science, the University of Tokyo
 */

#include "gbimpl.h"

bucket_t *g_connBuckets[3];
integer g_connMin[3], g_connMax[3];
itemp_t *g_vconnlist[3];

void ConnBucketInitial(bucket_t **buckets, integer nb, itemp_t **vlist, integer ni);
void ConnBucketFinalize(bucket_t **buckets, itemp_t **vlist, integer ni);
void ConnBucketReset(bucket_t *buckets, integer nb, itemp_t *vlist, integer ni);
void ConnBucketInsert(integer conn, integer vertex, bucket_t *buckets, integer *connMin, integer *connMax, itemp_t *vlist);
void ConnBucketDelete(integer conn, integer vertex, bucket_t *buckets, integer *connMin, integer *connMax, itemp_t *vlist);
void ConnBucketIncrease(integer conn, integer vertex, bucket_t *buckets, integer *connMin, integer *connMax, itemp_t *vlist);
integer FirstMax(int addset, int otherset, bucket_t **buckets, integer *connMin, integer *connMax, itemp_t **vlist);

void GB_MinMaxGreedyBisect() {
	integer i, j;

	g_timelimit = g_timeout + g_bi_start;
	for (i = 0; i < g_iter; i++) {
		for (j = 1; j <= *g_nv; j++) {
			(*g_belonglist)[j] = GB_PAR_L;
		}
		GB_MinMaxGreedyBisectSub();
		if (g_exitflag == GB_TRUE) {
			goto FOUND_EXIT;
		}
	}
FOUND_EXIT:
	*g_edgecut = g_bestedgecut;
	for (i = 1; i <= *g_nv; i++) {
		(*g_belonglist)[i] = g_bestbelonglist[i]; 
	}
}


void GB_MinMaxGreedyBisectSub() {
	const char FUNCNAME[] = "GB_MinMaxGreedyBisectSub(1)";

	integer seed1, seed2, addset, otherset, tobeadded, bestvertex;
	integer tempset, tempvertex, i;

	if (g_connBuckets[GB_PAR_L] == NULL || g_connBuckets[GB_PAR_R] == NULL 
		|| g_vconnlist[GB_PAR_L] == NULL || g_vconnlist[GB_PAR_R] == NULL || g_randomvlist == NULL) {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}
	/* Initialize start */
	ConnBucketReset(g_connBuckets[GB_PAR_L], *g_maxvdeg + 1, g_vconnlist[GB_PAR_L], *g_nv + 1);
	ConnBucketReset(g_connBuckets[GB_PAR_R], *g_maxvdeg + 1, g_vconnlist[GB_PAR_R], *g_nv + 1);
	g_connMin[GB_PAR_L] = g_connMin[GB_PAR_R] = *g_ne + 1;
	g_connMax[GB_PAR_L] = g_connMax[GB_PAR_R] = -1;
	/* Initialize end */

	/* Setup two seeds within range */
	if (g_seedL != g_seedU) {
		seed1 = GB_Random(g_seedL, g_seedU - g_seedL + 1);
		do {
			seed2 = GB_Random(g_seedL, g_seedU - g_seedL + 1);
		} while (seed1 == seed2);
	} else {
		seed1 = g_seedL;
		do {
			seed2 = GB_Random(1, *g_nv);
		} while (seed1 == seed2);
	}
	/* Generate random permutation */
	GB_RandomPermuteExclude(g_randomvlist, *g_nv, seed1, seed2);

	*g_edgecut = 0;
	addset = GB_PAR_L;
	otherset = GB_PAR_R;
	(*g_belonglist)[seed1] = addset;
	(*g_belonglist)[seed2] = otherset;
	tobeadded = *g_nv - 2;

	for (i = 3; i <= *g_nv; i++) {
		tempvertex = g_randomvlist[i];
		ConnBucketInsert(0, tempvertex, g_connBuckets[addset], &(g_connMin[addset]), &(g_connMax[addset]), g_vconnlist[addset]);
		ConnBucketInsert(0, tempvertex, g_connBuckets[otherset], &(g_connMin[otherset]), &(g_connMax[otherset]), g_vconnlist[otherset]);
		(*g_belonglist)[tempvertex] = GB_PAR_N;
	}

	/* Update seeds' adjacent vertices in buckets */
	for (i = 1; i <= (*g_adjlist)[seed1][0]; i++) {
		tempvertex = (*g_adjlist)[seed1][i];
		if ((*g_belonglist)[tempvertex] == GB_PAR_N) {
			ConnBucketIncrease(0, tempvertex, g_connBuckets[addset], &(g_connMin[addset]), &(g_connMax[addset]), g_vconnlist[addset]);
		}
	}
	for (i = 1; i <= (*g_adjlist)[seed2][0]; i++) {
		tempvertex = (*g_adjlist)[seed2][i];
		if (tempvertex == seed1) {
			*g_edgecut = 1;
		}
		if ((*g_belonglist)[tempvertex] == GB_PAR_N) {
			ConnBucketIncrease(0, tempvertex, g_connBuckets[otherset], &(g_connMin[otherset]), &(g_connMax[otherset]), g_vconnlist[otherset]);
		}
	}
	
	addset = GB_PAR_R;
	otherset = GB_PAR_L;
	while (tobeadded > 0) {
		tempset = addset;
		addset = otherset;
		otherset = tempset;
		
		bestvertex = FirstMax(addset, otherset, g_connBuckets, g_connMin, g_connMax, g_vconnlist);
		(*g_belonglist)[bestvertex] = addset;
		*g_edgecut += g_connMin[otherset];
		ConnBucketDelete(g_vconnlist[addset][bestvertex].value, bestvertex, g_connBuckets[addset], &(g_connMin[addset]), &(g_connMax[addset]), g_vconnlist[addset]);
		ConnBucketDelete(g_vconnlist[otherset][bestvertex].value, bestvertex, g_connBuckets[otherset], &(g_connMin[otherset]), &(g_connMax[otherset]), g_vconnlist[otherset]);
		for (i = 1; i <= (*g_adjlist)[bestvertex][0]; i++) {
			tempvertex = (*g_adjlist)[bestvertex][i];
			if ((*g_belonglist)[tempvertex] == GB_PAR_N) {
				ConnBucketIncrease(g_vconnlist[addset][tempvertex].value, tempvertex, g_connBuckets[addset], &(g_connMin[addset]), &(g_connMax[addset]), g_vconnlist[addset]);
			}
		}
		tobeadded--;
	}
	if (*g_edgecut < g_bestedgecut) {
		g_bestedgecut = *g_edgecut;
		for (i = 1; i <= *g_nv; i++) {
			g_bestbelonglist[i] = (*g_belonglist)[i];
		}
	}
	if (g_bestedgecut <= g_threshold || (g_timeout > 0 && clock() >= g_timelimit)) {
		g_exitflag = GB_TRUE;
	}
}

integer FirstMax(int addset, int otherset, bucket_t **buckets, integer *connMin, integer *connMax, itemp_t **vlist) {
	const char FUNCNAME[] = "FirstMax";
	integer min, max, vertex, i;
	item_t *curr;

	min = connMin[otherset];
	max = connMax[addset];
	
	vertex = 0;
	for (i = max; i >= 0; i--) {
		curr = buckets[addset][i].head;
		while (curr != NULL) {
			vertex = curr->index;
			if (vlist[otherset][vertex].value == min) {
				goto FOUND_EXIT;
			}
			curr = curr->next;
		}
	}

	if (vertex == 0) {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}
FOUND_EXIT:
	return vertex;
}

void ConnBucketInitial(bucket_t **buckets, integer nb, itemp_t **vlist, integer ni) {
	GB_BucketInitial(buckets, nb, vlist, ni, -1);
}

void ConnBucketFinalize(bucket_t **buckets, itemp_t **vlist, integer ni) {
	GB_BucketFinalize(buckets, vlist, ni);
}

void ConnBucketReset(bucket_t *buckets, integer nb, itemp_t *vlist, integer ni) {
	GB_BucketReset(buckets, nb, vlist, ni, -1);
}

void ConnBucketInsert(integer conn, integer vertex, bucket_t *buckets, integer *connMin, integer *connMax, itemp_t *vlist) {
	GB_BucketInsert(conn, vertex, buckets, connMin, connMax, vlist);
}

void ConnBucketDelete(integer conn, integer vertex, bucket_t *buckets, integer *connMin, integer *connMax, itemp_t *vlist) {
	GB_BucketDelete(conn, vertex, buckets, connMin, connMax, vlist);
}

void ConnBucketIncrease(integer conn, integer vertex, bucket_t *buckets, integer *connMin, integer *connMax, itemp_t *vlist) {
	GB_BucketDelete(conn, vertex, buckets, connMin, connMax, vlist);
	GB_BucketInsert(conn + 1, vertex, buckets, connMin, connMax, vlist);
}

